@@ -13,4 +13,5 @@ except PackageNotFoundError:  # pragma: no cover
     __version__ = "unknown"
 
 from layer_to_layer_pytorch.l2l import Layer2Layer
-from layer_to_layer_pytorch.types import Device, TensorOrTensorArray
+from layer_to_layer_pytorch.loss import L2LLoss
+from layer_to_layer_pytorch.types import Device, LossFn, TensorOrTensorArray
